/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "tstore.h"

uint64_t store_size;
int store_fd,hash_fd;
char *namebuf;
char *storename;
int maxused=0,maxsecused=0,overflows=0;
uint8_t *hash_map;

int zblock(uint32_t *buf)
{
  int i;
  for (i=0;i<128;++i)
    if (buf[i]) return 0;
  return 1;
}

int match(uint8_t *buf, uint64_t secnum)
{
  uint8_t buf2[512];

  switch (pread(store_fd,buf2,512,secnum*512)) {
  case -1:
    perror(storename);
    exit(1);
  default:
    fprintf(stderr,"%s: Incomplete pread\n",storename);
    exit(1);
  case 512:
    break;
  }
  return !memcmp(buf2,buf,512);
}

/* Basic principle as in put_hash() */
uint64_t put_block_in_bucket(uint8_t *bucket, uint8_t *buf,
			     uint16_t small_hash, int *max)
{
  int i;
  uint64_t r=0;
  int used=bucket[0];

  if (used>*max) *max=used;

  for (i=0;(i<used)&&(i<73);++i)
    if (GET_HASH(bucket+(6+7*i))==small_hash) {
      uint64_t secnum=GET_BNUM(bucket+(1+7*i));
      if (match(buf,secnum)) return secnum;
    }

  if (used<73) {
    switch (write(store_fd,buf,512)) {
    case -1:
      perror(storename);
      exit(1);
    default:
      fprintf(stderr,"%s: Incomplete write\n",storename);
      exit(1);
    case 512:
      break;
    }

    PUT_BNUM(bucket+(1+7*used),store_size);
    PUT_HASH(bucket+(6+7*used),small_hash);

    r=store_size++;

    if (max==&maxsecused) ++overflows;
  }

  if (used<255) ++used;
  if (used>*max) *max=used;
  bucket[0]=used;

  return r;
}

uint64_t put_block(uint8_t *buf, int useseq)
{
  if (zblock((void*)buf)) return 0;
  {
    uint64_t this_hash=hash_sector(buf);
    uint64_t bucket_number=this_hash>>16;
    uint16_t small_hash=this_hash;
    uint64_t r;

    if (useseq&&seqblock(buf)) return GET_BNUM(buf)+(1LL<<38);

    r=put_block_in_bucket(hash_map+512*bucket_number,buf,small_hash,&maxused);
    if (r) return r;
    bucket_number=pri_buckets+(this_hash%sec_buckets);
    r=put_block_in_bucket(hash_map+512*bucket_number,buf,small_hash,&maxsecused);
    if (r) return r;
  }

  fprintf(stderr,"Secondary hash overflow!\n");
  exit(1);
}

uint8_t pointers[8][512];
int usage[16];

void put_pointer(uint64_t p, int level)
{
  PUT_BNUM(pointers[level]+5*usage[level],p);
  usage[level]++;
  if (usage[level]==102) {
    put_pointer(put_block(pointers[level],1),level+1);
    usage[level]=0;
  }
}

uint64_t flush_pointers(int levels)
{
  int i;
  for (i=0;i<levels;++i)
    while(usage[i])
      put_pointer(0,i);
  return GET_BNUM(pointers[levels]);
}

int main(int argc, char ** argv)
{
  uint64_t filesize=0;
  uint8_t buf[512];
  int r;
  char formbuf[42];

  char *progbuf=getenv("TSTORE");
  if (progbuf) sprintf(formbuf,"%%%dlld",strlen(progbuf));

  memset(usage,0,sizeof(usage));
  memset(pointers,0,sizeof(pointers));

  storename=argv[1];

  if (argc != 3) {
    fprintf(stderr,"Uage: %s <store> <file>\n",argv[0]);
    return 1;
  }

  if (strlen(argv[2])>494) {
    fprintf(stderr,"%s: Name too long\n",argv[2]);
    return 1;
  }

  store_fd=open(argv[1],O_RDWR);
  if (store_fd==-1) {
    perror(argv[1]);
    return 1;
  }

  store_size=test_store_size(store_fd,1);
  store_error(store_size,argv[1]);
  lseek(store_fd,0,SEEK_END);

  namebuf=malloc(strlen(argv[1])+10);
  if (!namebuf) {
    perror("Out of memory");
    return 1;
  }

  sprintf(namebuf,"%s.idx",argv[1]);

  hash_fd=open(namebuf,O_RDWR,0666);
  if (hash_fd==-1) {
    perror(namebuf);
    return 1;
  }

  buckets=size_in_blocks(hash_fd);
  if (!buckets) {
    fprintf(stderr,"Invalid hash index (rebuild?)\n");
    return 1;
  }

  hash_map=mmap(NULL,buckets*512,PROT_READ|PROT_WRITE,MAP_SHARED,hash_fd,0);
  if (hash_map==MAP_FAILED) {
    perror("mmap");
    return 1;
  }

  init_hash();

  do {
    r=myread(0,buf,512);
    if (r==-1) {
      perror(argv[2]);
      return 1;
    }
    if (r) {
      filesize += r;
      put_pointer(put_block(buf,0),0);
      if (progbuf) sprintf(progbuf,formbuf,filesize);
    }
  } while(r==512);

  if (filesize) {
    uint64_t direntry;
    uint64_t block_directory=flush_pointers(filelevels(filesize));

    memset(buf,0,512);
    assert(pread(store_fd,buf,5,13)==5);
    PUT_BNUM(buf+5,block_directory);
    PUT_SIZE(buf+10,filesize);
    strcpy(buf+17,argv[2]);

    direntry=put_block(buf,0);
    PUT_BNUM(buf,store_size);
    PUT_BNUM(buf+5,direntry);
    assert(pwrite(store_fd,buf,10,8)==10);
  }

  hash_advice(overflows,maxused,maxsecused);

  return 0;
}
